"""
Tests for config models decorators
"""
from example.models import ExampleDecoratorConfig

from config_models import decorators
from tests.utils import CacheIsolationTestCase


@decorators.require_config(ExampleDecoratorConfig)
def decorated_fake_view():
    """
    Test function to wrap in our decorator.
    """
    return "success"


class RequireConfigTests(CacheIsolationTestCase):
    """
    Tests the require_config decorator.
    """

    def test_no_config(self):
        self.assertIs(decorators.HttpResponseNotFound, type(decorated_fake_view()))

    def test_config_enabled(self):
        ExampleDecoratorConfig.objects.create(enabled=True)
        self.assertEqual(decorated_fake_view(), "success")

    def test_config_disabled(self):
        ExampleDecoratorConfig.objects.create(enabled=False)
        self.assertIs(decorators.HttpResponseNotFound, type(decorated_fake_view()))



"""
Tests of ConfigurationModel
"""
from unittest import mock

import ddt
from django.contrib.auth import get_user_model
from edx_django_utils.cache.utils import CachedResponse
from example.models import (ExampleConfig, ExampleKeyedConfig,
                            ManyToManyExampleConfig)
from freezegun import freeze_time
from rest_framework.test import APIRequestFactory

from config_models.views import ConfigurationModelCurrentAPIView

from .utils import CacheIsolationTestCase

User = get_user_model()


class ConfigurationModelTests(CacheIsolationTestCase):
    """
    Tests of ConfigurationModel
    """
    def setUp(self):
        super().setUp()
        self.user = User()
        self.user.save()

    def test_cache_key_name(self):
        self.assertEqual(
            ExampleConfig.cache_key_name(),
            'configuration/ExampleConfig/current'
        )

    def test_no_config_empty_cache(self):
        # First time reads from the database
        with self.assertNumQueries(1):
            current = ExampleConfig.current()
            self.assertEqual(current.int_field, 10)
            self.assertEqual(current.string_field, '')

        # Follow-on request reads from cache with no database request.
        with self.assertNumQueries(0):
            cached_current = ExampleConfig.current()
            self.assertEqual(cached_current.int_field, 10)
            self.assertEqual(cached_current.string_field, '')

    @mock.patch('config_models.models.TieredCache.get_cached_response')
    def test_config_with_cached_response_value_none(self, mock_cached_response):
        mock_cached_response.return_value = CachedResponse(is_found=True, key='test-key', value=None)
        # First time reads from the database
        with self.assertNumQueries(1):
            current = ExampleConfig.current()
            self.assertEqual(current.int_field, 10)
            self.assertEqual(current.string_field, '')

        # Follow-on request reads from database instead of cache as cache value will be None.
        with self.assertNumQueries(1):
            cached_current = ExampleConfig.current()
            self.assertEqual(cached_current.int_field, 10)
            self.assertEqual(cached_current.string_field, '')

    def test_config_ordering(self):
        with freeze_time('2012-01-01'):
            first = ExampleConfig(changed_by=self.user)
            first.string_field = 'first'
            first.save()

        second = ExampleConfig(changed_by=self.user)
        second.string_field = 'second'
        second.save()

        self.assertEqual(ExampleConfig.current().string_field, 'second')

    def test_cache_set(self):
        # Nothing is cached, so we take a database hit for this.
        with self.assertNumQueries(1):
            default = ExampleConfig.current()
            self.assertEqual(default.string_field, '')

        first = ExampleConfig(changed_by=self.user)
        first.string_field = 'first'
        first.save()

        # The save() call above should have invalidated the cache, so we expect
        # to see a query to get the current config value.
        with self.assertNumQueries(1):
            current = ExampleConfig.current()
            self.assertEqual(current.string_field, 'first')

    def test_active_annotation(self):
        with freeze_time('2012-01-01'):
            ExampleConfig.objects.create(string_field='first')

        ExampleConfig.objects.create(string_field='second')

        rows = ExampleConfig.objects.with_active_flag().order_by('-change_date')
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0].string_field, 'second')
        self.assertEqual(rows[0].is_active, True)
        self.assertEqual(rows[1].string_field, 'first')
        self.assertEqual(rows[1].is_active, False)

    def test_keyed_active_annotation(self):
        with freeze_time('2012-01-01'):
            ExampleKeyedConfig.objects.create(left='left', right='right', user=self.user, string_field='first')

        ExampleKeyedConfig.objects.create(left='left', right='right', user=self.user, string_field='second')

        rows = ExampleKeyedConfig.objects.with_active_flag().order_by('-change_date')
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0].string_field, 'second')
        self.assertEqual(rows[0].is_active, True)
        self.assertEqual(rows[1].string_field, 'first')
        self.assertEqual(rows[1].is_active, False)

    def test_always_insert(self):
        config = ExampleConfig(changed_by=self.user, string_field='first')
        config.save()
        config.string_field = 'second'
        config.save()

        self.assertEqual(2, ExampleConfig.objects.all().count())

    def test_equality(self):
        config = ExampleConfig(changed_by=self.user, string_field='first')
        config.save()

        self.assertTrue(ExampleConfig.equal_to_current({"string_field": "first"}))
        self.assertTrue(ExampleConfig.equal_to_current({"string_field": "first", "enabled": False}))
        self.assertTrue(ExampleConfig.equal_to_current({"string_field": "first", "int_field": 10}))

        self.assertFalse(ExampleConfig.equal_to_current({"string_field": "first", "enabled": True}))
        self.assertFalse(ExampleConfig.equal_to_current({"string_field": "first", "int_field": 20}))
        self.assertFalse(ExampleConfig.equal_to_current({"string_field": "second"}))

        self.assertFalse(ExampleConfig.equal_to_current({}))

    def test_equality_custom_fields_to_ignore(self):
        config = ExampleConfig(changed_by=self.user, string_field='first')
        config.save()

        # id, change_date, and changed_by will all be different for a newly created entry
        self.assertTrue(ExampleConfig.equal_to_current({"string_field": "first"}))
        self.assertFalse(
            ExampleConfig.equal_to_current({"string_field": "first"}, fields_to_ignore=("change_date", "changed_by"))
        )
        self.assertFalse(
            ExampleConfig.equal_to_current({"string_field": "first"}, fields_to_ignore=("id", "changed_by"))
        )
        self.assertFalse(
            ExampleConfig.equal_to_current({"string_field": "first"}, fields_to_ignore=("change_date", "id"))
        )

        # Test the ability to ignore a different field ("int_field").
        self.assertFalse(ExampleConfig.equal_to_current({"string_field": "first", "int_field": 20}))
        self.assertTrue(
            ExampleConfig.equal_to_current(
                {"string_field": "first", "int_field": 20},
                fields_to_ignore=("id", "change_date", "changed_by", "int_field")
            )
        )

    def test_equality_ignores_many_to_many(self):
        config = ManyToManyExampleConfig(changed_by=self.user, string_field='first')
        config.save()

        second_user = User(username="second_user")
        second_user.save()
        config.many_user_field.add(second_user)
        config.save()

        # The many-to-many field is ignored in comparison.
        self.assertTrue(
            ManyToManyExampleConfig.equal_to_current({"string_field": "first", "many_user_field": "removed"})
        )


@ddt.ddt
class KeyedConfigurationModelTests(CacheIsolationTestCase):
    """
    Tests for ``ConfigurationModels`` with keyed configuration.
    """
    def setUp(self):
        super().setUp()
        self.user = User()
        self.user.save()

    @ddt.data(('a', 'b'), ('c', 'd'))
    @ddt.unpack
    def test_cache_key_name(self, left, right):
        self.assertEqual(
            ExampleKeyedConfig.cache_key_name(left, right, self.user),
            f'configuration/ExampleKeyedConfig/current/{left},{right},{self.user}'
        )

    @ddt.data(
        ((), 'left,right,user'),
        (('left', 'right'), 'left,right'),
        (('left', ), 'left')
    )
    @ddt.unpack
    def test_key_values_cache_key_name(self, args, expected_key):
        self.assertEqual(
            ExampleKeyedConfig.key_values_cache_key_name(*args),
            f'configuration/ExampleKeyedConfig/key_values/{expected_key}')

    @ddt.data(('a', 'b'), ('c', 'd'))
    @ddt.unpack
    def test_no_config_empty_cache(self, left, right):
        # First time, it comes from the database.
        with self.assertNumQueries(1):
            current = ExampleKeyedConfig.current(left, right, self.user)
            self.assertEqual(current.int_field, 10)
            self.assertEqual(current.string_field, '')

        # Second time, it should come from cache.
        with self.assertNumQueries(0):
            current = ExampleKeyedConfig.current(left, right, self.user)
            self.assertEqual(current.int_field, 10)
            self.assertEqual(current.string_field, '')

    def test_config_ordering(self):
        with freeze_time('2012-01-01'):
            ExampleKeyedConfig(
                changed_by=self.user,
                left='left_a',
                right='right_a',
                string_field='first_a',
                user=self.user,
            ).save()

            ExampleKeyedConfig(
                changed_by=self.user,
                left='left_b',
                right='right_b',
                string_field='first_b',
                user=self.user,
            ).save()

        ExampleKeyedConfig(
            changed_by=self.user,
            left='left_a',
            right='right_a',
            string_field='second_a',
            user=self.user,
        ).save()
        ExampleKeyedConfig(
            changed_by=self.user,
            left='left_b',
            right='right_b',
            string_field='second_b',
            user=self.user,
        ).save()

        self.assertEqual(ExampleKeyedConfig.current('left_a', 'right_a', self.user).string_field, 'second_a')
        self.assertEqual(ExampleKeyedConfig.current('left_b', 'right_b', self.user).string_field, 'second_b')

    def test_cache_set(self):
        with self.assertNumQueries(1):
            default = ExampleConfig.current('left', 'right', self.user)
            self.assertEqual(default.string_field, '')

        first = ExampleKeyedConfig(
            changed_by=self.user,
            left='left',
            right='right',
            user=self.user,
            string_field='first',
        )
        first.save()

        # Cache is invalidated by the save above, so another query is made.
        with self.assertNumQueries(1):
            ExampleKeyedConfig.current('left', 'right', self.user)

    def test_key_values(self):
        with freeze_time('2012-01-01'):
            ExampleKeyedConfig(left='left_a', right='right_a', user=self.user, changed_by=self.user).save()
            ExampleKeyedConfig(left='left_b', right='right_b', user=self.user, changed_by=self.user).save()

        ExampleKeyedConfig(left='left_a', right='right_a', user=self.user, changed_by=self.user).save()
        ExampleKeyedConfig(left='left_b', right='right_b', user=self.user, changed_by=self.user).save()

        unique_key_pairs = ExampleKeyedConfig.key_values()
        self.assertEqual(len(unique_key_pairs), 2)
        self.assertEqual(
            set(unique_key_pairs),
            {('left_a', 'right_a', self.user.id), ('left_b', 'right_b', self.user.id)}
        )
        unique_left_keys = ExampleKeyedConfig.key_values('left', flat=True)
        self.assertEqual(len(unique_left_keys), 2)
        self.assertEqual(set(unique_left_keys), {'left_a', 'left_b'})

    def test_key_string_values(self):
        """ Ensure str() vs unicode() doesn't cause duplicate cache entries """
        ExampleKeyedConfig(
            left='left',
            right='\N{RIGHT ANGLE BRACKET}\N{SNOWMAN}',
            enabled=True,
            int_field=10,
            user=self.user,
            changed_by=self.user
        ).save()

        entry = ExampleKeyedConfig.current('left', '\N{RIGHT ANGLE BRACKET}\N{SNOWMAN}', self.user)
        self.assertEqual(entry.int_field, 10)

        entry = ExampleKeyedConfig.current('left', '\N{RIGHT ANGLE BRACKET}\N{SNOWMAN}', self.user)
        self.assertEqual(entry.int_field, 10)

    def test_current_set(self):
        with freeze_time('2012-01-01'):
            ExampleKeyedConfig(left='left_a', right='right_a', int_field=0, user=self.user, changed_by=self.user).save()
            ExampleKeyedConfig(left='left_b', right='right_b', int_field=0, user=self.user, changed_by=self.user).save()

        ExampleKeyedConfig(left='left_a', right='right_a', int_field=1, user=self.user, changed_by=self.user).save()
        ExampleKeyedConfig(left='left_b', right='right_b', int_field=2, user=self.user, changed_by=self.user).save()

        queryset = ExampleKeyedConfig.objects.current_set()
        self.assertEqual(len(queryset.all()), 2)
        self.assertEqual(
            set(queryset.order_by('int_field').values_list('int_field', flat=True)),
            {1, 2}
        )

    def test_active_annotation(self):
        with freeze_time('2012-01-01'):
            ExampleKeyedConfig.objects.create(left='left_a', right='right_a', user=self.user, string_field='first')
            ExampleKeyedConfig.objects.create(left='left_b', right='right_b', user=self.user, string_field='first')

        ExampleKeyedConfig.objects.create(left='left_a', right='right_a', user=self.user, string_field='second')

        rows = ExampleKeyedConfig.objects.with_active_flag()
        self.assertEqual(len(rows), 3)
        for row in rows:
            if row.left == 'left_a':
                self.assertEqual(row.is_active, row.string_field == 'second')
            else:
                self.assertEqual(row.left, 'left_b')
                self.assertEqual(row.string_field, 'first')
                self.assertEqual(row.is_active, True)

    def test_equality(self):
        config1 = ExampleKeyedConfig(left='left_a', right='right_a', int_field=1, user=self.user, changed_by=self.user)
        config1.save()

        config2 = ExampleKeyedConfig(
            left='left_b',
            right='right_b',
            int_field=2,
            user=self.user,
            changed_by=self.user,
            enabled=True,
        )
        config2.save()

        config3 = ExampleKeyedConfig(left='left_c', user=self.user, changed_by=self.user)
        config3.save()

        self.assertTrue(
            ExampleKeyedConfig.equal_to_current({
                "left": "left_a",
                "right": "right_a",
                "int_field": 1,
                "user": self.user
            })
        )
        self.assertTrue(
            ExampleKeyedConfig.equal_to_current({
                "left": "left_b",
                "right": "right_b",
                "int_field": 2,
                "user": self.user,
                "enabled": True
            })
        )
        self.assertTrue(
            ExampleKeyedConfig.equal_to_current({"left": "left_c", "user": self.user})
        )

        self.assertFalse(
            ExampleKeyedConfig.equal_to_current(
                {"left": "left_a", "right": "right_a", "user": self.user, "int_field": 1, "string_field": "foo"}
            )
        )
        self.assertFalse(
            ExampleKeyedConfig.equal_to_current({"left": "left_a", "user": self.user, "int_field": 1})
        )
        self.assertFalse(
            ExampleKeyedConfig.equal_to_current({
                "left": "left_b",
                "right": "right_b",
                "user": self.user,
                "int_field": 2,
            })
        )
        self.assertFalse(
            ExampleKeyedConfig.equal_to_current({"left": "left_c", "user": self.user, "int_field": 11})
        )

        self.assertFalse(ExampleKeyedConfig.equal_to_current({"user": self.user}))


@ddt.ddt
class ConfigurationModelAPITests(CacheIsolationTestCase):
    """
    Tests for the configuration model API.
    """
    def setUp(self):
        super().setUp()
        self.factory = APIRequestFactory()
        self.user = User.objects.create_user(
            username='test_user',
            email='test_user@example.com',
            password='test_pass',
        )
        self.user.is_superuser = True
        self.user.save()
        self.current_view = ConfigurationModelCurrentAPIView.as_view(model=ExampleConfig)

    def test_insert(self):
        self.assertEqual("", ExampleConfig.current().string_field)

        request = self.factory.post('/config/ExampleConfig', {"string_field": "string_value"})
        request.user = self.user
        __ = self.current_view(request)

        self.assertEqual("string_value", ExampleConfig.current().string_field)
        self.assertEqual(self.user, ExampleConfig.current().changed_by)

    def test_multiple_inserts(self):
        for i in range(3):
            self.assertEqual(i, ExampleConfig.objects.all().count())

            request = self.factory.post('/config/ExampleConfig', {"string_field": str(i)})
            request.user = self.user
            response = self.current_view(request)
            self.assertEqual(201, response.status_code)

            self.assertEqual(i + 1, ExampleConfig.objects.all().count())
            self.assertEqual(str(i), ExampleConfig.current().string_field)

    def test_get_current(self):
        request = self.factory.get('/config/ExampleConfig')
        request.user = self.user
        response = self.current_view(request)
        self.assertEqual('', response.data['string_field'])
        self.assertEqual(10, response.data['int_field'])
        self.assertEqual(None, response.data['changed_by'])
        self.assertEqual(False, response.data['enabled'])
        self.assertEqual(None, response.data['change_date'])

        ExampleConfig(string_field='string_value', int_field=20).save()

        response = self.current_view(request)
        self.assertEqual('string_value', response.data['string_field'])
        self.assertEqual(20, response.data['int_field'])

    @ddt.data(
        ('get', [], 200),
        ('post', [{'string_field': 'string_value', 'int_field': 10}], 201),
    )
    @ddt.unpack
    def test_permissions(self, method, args, status_code):
        request = getattr(self.factory, method)('/config/ExampleConfig', *args)

        request.user = User.objects.create_user(
            username='no-perms',
            email='no-perms@example.com',
            password='no-perms',
        )
        response = self.current_view(request)
        self.assertEqual(403, response.status_code)

        request.user = self.user
        response = self.current_view(request)
        self.assertEqual(status_code, response.status_code)



"""
Django Model baseclass for database-backed configuration.
"""


from django.conf import settings
# The following import exists for backwards compatibility (because a number of
# library users assume config_models.models.cache is importable), but
# ConfigModels will now ignore the custom 'configuration' cache setting and just
# use TieredCache, which will make use of a local request cache + the default
# Django cache.
from django.core.cache import cache  # pylint: disable=unused-import
from django.db import models
from django.utils.translation import gettext_lazy as _
from edx_django_utils.cache.utils import TieredCache
from rest_framework.utils import model_meta


class ConfigurationModelManager(models.Manager):
    """
    Query manager for ConfigurationModel
    """
    def _current_ids_subquery(self):
        """
        Internal helper method to return an SQL string that will get the IDs of
        all the current entries (i.e. the most recent entry for each unique set
        of key values). Only useful if KEY_FIELDS is set.
        """
        return self.values(*self.model.KEY_FIELDS).annotate(max=models.Max('pk')).values('max')

    def current_set(self):
        """
        A queryset for the active configuration entries only. Only useful if KEY_FIELDS is set.

        Active means the means recent entries for each unique combination of keys. It does not
        necessaryily mean enbled.
        """
        assert self.model.KEY_FIELDS != (), "Just use model.current() if there are no KEY_FIELDS"
        return self.get_queryset().filter(
            pk__in=self._current_ids_subquery()
        ).annotate(
            is_active=models.Value(1, output_field=models.IntegerField())
        )

    def with_active_flag(self):
        """
        A query set where each result is annotated with an 'is_active' field that indicates
        if it's the most recent entry for that combination of keys.
        """
        if self.model.KEY_FIELDS:
            return self.get_queryset().annotate(
                is_active=models.ExpressionWrapper(
                    models.Q(pk__in=self._current_ids_subquery()),
                    output_field=models.IntegerField(),
                )
            )
        return self.get_queryset().annotate(
            is_active=models.ExpressionWrapper(
                models.Q(pk=self.model.current().pk),
                output_field=models.IntegerField(),
            )
        )


class ConfigurationModel(models.Model):
    """
    Abstract base class for model-based configuration

    Properties:
        cache_timeout (int): The number of seconds that this configuration
            should be cached
    """

    class Meta:
        abstract = True
        ordering = ("-change_date", )

    objects = ConfigurationModelManager()

    KEY_FIELDS = ()

    # The number of seconds
    cache_timeout = 600

    change_date = models.DateTimeField(auto_now_add=True, verbose_name=_("Change date"))
    changed_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        null=True,
        on_delete=models.PROTECT,
        # Translators: this label indicates the name of the user who made this change:
        verbose_name=_("Changed by"),
    )
    enabled = models.BooleanField(default=False, verbose_name=_("Enabled"))

    def save(self, force_insert=False, force_update=False, using=None,
             update_fields=None):
        """
        Clear the cached value when saving a new configuration entry
        """
        # Always create a new entry, instead of updating an existing model
        self.pk = None  # pylint: disable=invalid-name
        super().save(
            force_insert,
            force_update,
            using,
            update_fields
        )
        TieredCache.delete_all_tiers(self.cache_key_name(*[getattr(self, key) for key in self.KEY_FIELDS]))
        if self.KEY_FIELDS:
            TieredCache.delete_all_tiers(self.key_values_cache_key_name())

    @classmethod
    def cache_key_name(cls, *args):
        """Return the name of the key to use to cache the current configuration"""
        if cls.KEY_FIELDS != ():  # pylint: disable=use-implicit-booleaness-not-comparison
            if len(args) != len(cls.KEY_FIELDS):
                raise TypeError(
                    f"cache_key_name() takes exactly {len(cls.KEY_FIELDS)} arguments ({len(args)} given)"
                )
            return f"configuration/{cls.__name__}/current/{','.join(str(arg) for arg in args)}"
        else:
            return f'configuration/{cls.__name__}/current'

    @classmethod
    def current(cls, *args):
        """
        Return the active configuration entry, either from cache,
        from the database, or by creating a new empty entry (which is not
        persisted).
        """
        cache_key = cls.cache_key_name(*args)
        cached_response = TieredCache.get_cached_response(cache_key)
        if cached_response.is_found and cached_response.value is not None:
            return cached_response.value

        key_dict = dict(zip(cls.KEY_FIELDS, args))
        try:
            current = cls.objects.filter(**key_dict).order_by('-change_date')[0]
        except IndexError:
            current = cls(**key_dict)

        TieredCache.set_all_tiers(cache_key, current, cls.cache_timeout)
        return current

    @classmethod
    def is_enabled(cls, *key_fields):
        """
        Returns True if this feature is configured as enabled, else False.

        Arguments:
            key_fields: The positional arguments are the KEY_FIELDS used to identify the
                configuration to be checked.
        """
        return cls.current(*key_fields).enabled

    @classmethod
    def key_values_cache_key_name(cls, *key_fields):
        """ Key for fetching unique key values from the cache """
        key_fields = key_fields or cls.KEY_FIELDS
        return f"configuration/{cls.__name__}/key_values/{','.join(key_fields)}"

    @classmethod
    def key_values(cls, *key_fields, **kwargs):
        """
        Get the set of unique values in the configuration table for the given
        key[s]. Calling cls.current(*value) for each value in the resulting
        list should always produce an entry, though any such entry may have
        enabled=False.

        Arguments:
            key_fields: The positional arguments are the KEY_FIELDS to return. For example if
                you had a course embargo configuration where each entry was keyed on (country,
                course), then you might want to know "What countries have embargoes configured?"
                with cls.key_values('country'), or "Which courses have country restrictions?"
                with cls.key_values('course'). You can also leave this unspecified for the
                default, which returns the distinct combinations of all keys.
            flat: If you pass flat=True as a kwarg, it has the same effect as in Django's
                'values_list' method: Instead of returning a list of lists, you'll get one list
                of values. This makes sense to use whenever there is only one key being queried.

        Return value:
            List of lists of each combination of keys found in the database.
            e.g. [("Italy", "course-v1:SomeX+some+2015"), ...] for the course embargo example
        """
        flat = kwargs.pop('flat', False)
        assert not kwargs, "'flat' is the only kwarg accepted"
        key_fields = key_fields or cls.KEY_FIELDS
        cache_key = cls.key_values_cache_key_name(*key_fields)
        cached_response = TieredCache.get_cached_response(cache_key)
        if cached_response.is_found:
            return cached_response.value

        values = list(cls.objects.values_list(*key_fields, flat=flat).order_by().distinct())
        TieredCache.set_all_tiers(cache_key, values, cls.cache_timeout)
        return values

    def fields_equal(self, instance, fields_to_ignore=("id", "change_date", "changed_by")):
        """
        Compares this instance's fields to the supplied instance to test for equality.
        This will ignore any fields in `fields_to_ignore`.

        Note that this method ignores many-to-many fields.

        Args:
            instance: the model instance to compare
            fields_to_ignore: List of fields that should not be compared for equality. By default
            includes `id`, `change_date`, and `changed_by`.

        Returns: True if the checked fields are all equivalent, else False
        """
        for field in self._meta.get_fields():
            if not field.many_to_many and field.name not in fields_to_ignore:
                if getattr(instance, field.name) != getattr(self, field.name):
                    return False

        return True

    @classmethod
    def equal_to_current(cls, json, fields_to_ignore=("id", "change_date", "changed_by")):
        """
        Compares for equality this instance to a model instance constructed from the supplied JSON.
        This will ignore any fields in `fields_to_ignore`.

        Note that this method cannot handle fields with many-to-many associations, as those can only
        be set on a saved model instance (and saving the model instance will create a new entry).
        All many-to-many field entries will be removed before the equality comparison is done.

        Args:
            json: json representing an entry to compare
            fields_to_ignore: List of fields that should not be compared for equality. By default
            includes `id`, `change_date`, and `changed_by`.

        Returns: True if the checked fields are all equivalent, else False
        """

        # Remove many-to-many relationships from json.
        # They require an instance to be already saved.
        info = model_meta.get_field_info(cls)
        for field_name, relation_info in info.relations.items():
            if relation_info.to_many and (field_name in json):
                json.pop(field_name)

        new_instance = cls(**json)
        key_field_args = tuple(getattr(new_instance, key) for key in cls.KEY_FIELDS)
        current = cls.current(*key_field_args)
        # If current.id is None, no entry actually existed and the "current" method created it.
        if current.id is not None:
            return current.fields_equal(new_instance, fields_to_ignore)

        return False



"""
ConfigurationModel Admin Module Test Cases
"""
from unittest.mock import patch

from django.contrib.admin.sites import AdminSite
from django.contrib.auth import get_user_model
from django.contrib.messages.storage.fallback import FallbackStorage
from django.http import HttpRequest
from django.test import TestCase
from example.models import ExampleKeyedConfig

from config_models import admin
from config_models.models import ConfigurationModel

User = get_user_model()


class AdminTestCaseMixin:
    """
    Provide a request factory method.
    """

    def get_request(self):
        request = HttpRequest()
        request.session = "session"
        request._messages = FallbackStorage(request)  # pylint: disable=protected-access
        return request


class AdminTestCase(TestCase, AdminTestCaseMixin):
    """
    Test Case module for ConfigurationModel Admin
    """

    def setUp(self):
        super().setUp()
        self.conf_admin = admin.ConfigurationModelAdmin(ConfigurationModel, AdminSite())

    def test_default_fields(self):
        """
        Test: checking fields
        """
        request = self.get_request()
        self.assertEqual(
            list(self.conf_admin.get_form(request).base_fields), ["enabled"]
        )


class KeyedAdminTestCase(TestCase, AdminTestCaseMixin):
    """
    Test case module for KeyedConfigurationModelAdmin.
    """

    def get_edit_link(self):
        """
        Return an edit link from a KeyedConfigurationModelAdmin. Patch the `reverse`
        and `_` methods to modify the return value.
        """
        conf_admin = admin.KeyedConfigurationModelAdmin(ExampleKeyedConfig, AdminSite())
        request = self.get_request()
        ExampleKeyedConfig.objects.create(user=User.objects.create())
        config = conf_admin.get_queryset(request)[0]
        return conf_admin.edit_link(config)

    def test_edit_link(self):
        with patch.object(admin, "reverse", return_value="http://google.com"):
            self.assertEqual(
                '<a href="http://google.com?source=1">Update</a>', self.get_edit_link(),
            )

    def test_edit_link_xss_url(self):
        with patch.object(
                admin, "reverse", return_value='"><script>console.log("xss!")</script>'
        ):
            edit_link = self.get_edit_link()

        self.assertNotIn(
            "<script>", edit_link,
        )
        self.assertIn(
            "&lt;script&gt;", edit_link,
        )

    def test_edit_link_xss_translation(self):
        with patch.object(admin, "reverse", return_value=""):
            with patch.object(
                    admin, "_", return_value='<script>console.log("xss!")</script>'
            ):
                edit_link = self.get_edit_link()

        self.assertNotIn(
            "<script>", edit_link,
        )
        self.assertIn(
            "&lt;script&gt;", edit_link,
        )



"""
URLs for config_models.
"""

from django.urls import re_path
from django.views.generic import TemplateView

urlpatterns = [
    re_path(r'', TemplateView.as_view(template_name="config_models/base.html")),
]



"""
API view to allow manipulation of configuration models.
"""


from django.db import transaction
from rest_framework.authentication import SessionAuthentication
from rest_framework.generics import CreateAPIView, RetrieveAPIView
from rest_framework.permissions import DjangoModelPermissions

from config_models.utils import get_serializer_class


class ReadableOnlyByAuthors(DjangoModelPermissions):
    """Only allow access by users with `add` permissions on the model."""
    perms_map = DjangoModelPermissions.perms_map.copy()
    perms_map['GET'] = perms_map['OPTIONS'] = perms_map['HEAD'] = perms_map['POST']


class AtomicMixin:
    """Mixin to provide atomic transaction for as_view."""
    @classmethod
    def create_atomic_wrapper(cls, wrapped_func):
        """Returns a wrapped function."""
        def _create_atomic_wrapper(*args, **kwargs):
            """Actual wrapper."""
            # When a view call fails due to a permissions error, it raises an exception.
            # An uncaught exception breaks the DB transaction for any following DB operations
            # unless it's wrapped in a atomic() decorator or context manager.
            with transaction.atomic():
                return wrapped_func(*args, **kwargs)

        return _create_atomic_wrapper

    @classmethod
    def as_view(cls, **initkwargs):
        """Overrides as_view to add atomic transaction."""
        view = super().as_view(**initkwargs)
        return cls.create_atomic_wrapper(view)


class ConfigurationModelCurrentAPIView(AtomicMixin, CreateAPIView, RetrieveAPIView):
    """
    This view allows an authenticated user with the appropriate model permissions
    to read and write the current configuration for the specified `model`.

    Like other APIViews, you can use this by using a url pattern similar to the following::

        url(r'config/example_config$', ConfigurationModelCurrentAPIView.as_view(model=ExampleConfig))
    """
    authentication_classes = (SessionAuthentication,)
    permission_classes = (ReadableOnlyByAuthors,)
    model = None

    def get_queryset(self):
        return self.model.objects.all()

    def get_object(self):
        # Return the currently active configuration
        return self.model.current()

    def get_serializer_class(self):
        if self.serializer_class is None:
            self.serializer_class = get_serializer_class(self.model)

        return self.serializer_class

    def perform_create(self, serializer):
        # Set the requesting user as the one who is updating the configuration
        serializer.save(changed_by=self.request.user)



"""
Override the submit_row template tag to remove all save buttons from the
admin dashboard change view if the context has readonly marked in it.
"""
from django.contrib.admin.templatetags.admin_modify import register
from django.contrib.admin.templatetags.admin_modify import submit_row as original_submit_row


@register.inclusion_tag('admin/submit_line.html', takes_context=True)
def submit_row(context):
    """
    Overrides 'django.contrib.admin.templatetags.admin_modify.submit_row'.

    Manipulates the context going into that function by hiding all of the buttons
    in the submit row if the key `readonly` is set in the context.
    """
    ctx = original_submit_row(context)

    if context.get('readonly', False):
        ctx.update({
            'show_delete_link': False,
            'show_save_as_new': False,
            'show_save_and_add_another': False,
            'show_save_and_continue': False,
            'show_save': False,
        })

    return ctx



"""
Populates a ConfigurationModel by deserializing JSON data contained in a file.
"""
import os

from django.core.management.base import BaseCommand, CommandError
from django.utils.translation import gettext_lazy as _

from config_models.utils import deserialize_json


class Command(BaseCommand):
    """
    This command will deserialize the JSON data in the supplied file to populate
    a ConfigurationModel. Note that this will add new entries to the model, but it
    will not delete any entries (ConfigurationModel entries are read-only).
    """
    help = """
    Populates a ConfigurationModel by deserializing the supplied JSON.

    JSON should be in a file, with the following format:

    { "model": "config_models.ExampleConfigurationModel",
      "data":
        [
          { "enabled": True,
            "color": "black"
            ...
          },
          { "enabled": False,
            "color": "yellow"
            ...
          },
          ...
        ]
    }

    A username corresponding to an existing user must be specified to indicate who
    is executing the command.

        $ ... populate_model -f path/to/file.json -u username
    """

    def add_arguments(self, parser):
        parser.add_argument(
            '-f',
            '--file',
            metavar='JSON_FILE',
            dest='file',
            default=False,
            help='JSON file to import ConfigurationModel data'
        )

        parser.add_argument(
            '-u',
            '--username',
            metavar='USERNAME',
            dest='username',
            default=False,
            help='username to specify who is executing the command'
        )

    def handle(self, *args, **options):
        if 'file' not in options or not options['file']:
            raise CommandError(_("A file containing JSON must be specified."))

        if 'username' not in options or not options['username']:
            raise CommandError(_("A valid username must be specified."))

        json_file = options['file']
        if not os.path.exists(json_file):
            raise CommandError(_("File {0} does not exist").format(json_file))

        self.stdout.write(_("Importing JSON data from file {0}").format(json_file))
        with open(json_file, "rb") as data:
            created_entries = deserialize_json(data, options['username'])
            self.stdout.write(_("Import complete, {0} new entries created").format(created_entries))



"""
Tests of the populate_model management command and its helper utils.deserialize_json method.
"""


import io
import os.path
import textwrap

from django.contrib.auth import get_user_model
from django.core.management.base import CommandError
from django.utils import timezone
from example.models import ExampleDeserializeConfig

from config_models.management.commands import populate_model
from config_models.utils import deserialize_json
from tests.utils import CacheIsolationTestCase

User = get_user_model()


class DeserializeJSONTests(CacheIsolationTestCase):
    """
    Tests of deserializing the JSON representation of ConfigurationModels.
    """
    def setUp(self):
        super().setUp()
        self.test_username = 'test_worker'
        User.objects.create_user(username=self.test_username)
        self.fixture_path = os.path.join(os.path.dirname(__file__), 'data', 'data.json')

    def test_deserialize_models(self):
        """
        Tests the "happy path", where 2 instances of the test model should be created.
        A valid username is supplied for the operation.
        """
        start_date = timezone.now()
        with open(self.fixture_path, "rb") as data:
            entries_created = deserialize_json(data, self.test_username)
            self.assertEqual(2, entries_created)

        self.assertEqual(2, ExampleDeserializeConfig.objects.count())

        betty = ExampleDeserializeConfig.current('betty')
        self.assertTrue(betty.enabled)
        self.assertEqual(5, betty.int_field)
        self.assertGreater(betty.change_date, start_date)
        self.assertEqual(self.test_username, betty.changed_by.username)

        fred = ExampleDeserializeConfig.current('fred')
        self.assertFalse(fred.enabled)
        self.assertEqual(10, fred.int_field)
        self.assertGreater(fred.change_date, start_date)
        self.assertEqual(self.test_username, fred.changed_by.username)

    def test_existing_entries_not_removed(self):
        """
        Any existing configuration model entries are retained
        (though they may be come history)-- deserialize_json is purely additive.
        """
        ExampleDeserializeConfig(name="fred", enabled=True).save()
        ExampleDeserializeConfig(name="barney", int_field=200).save()

        with open(self.fixture_path, "rb") as data:
            entries_created = deserialize_json(data, self.test_username)
            self.assertEqual(2, entries_created)

        self.assertEqual(4, ExampleDeserializeConfig.objects.count())
        self.assertEqual(3, len(ExampleDeserializeConfig.objects.current_set()))

        self.assertEqual(5, ExampleDeserializeConfig.current('betty').int_field)
        self.assertEqual(200, ExampleDeserializeConfig.current('barney').int_field)

        # The JSON file changes "enabled" to False for Fred.
        fred = ExampleDeserializeConfig.current('fred')
        self.assertFalse(fred.enabled)

    def test_duplicate_entries_not_made(self):
        """
        If there is no change in an entry (besides changed_by and change_date),
        a new entry is not made.
        """
        with open(self.fixture_path, "rb") as data:
            entries_created = deserialize_json(data, self.test_username)
            self.assertEqual(2, entries_created)

        with open(self.fixture_path, "rb") as data:
            entries_created = deserialize_json(data, self.test_username)
            self.assertEqual(0, entries_created)

        # Importing twice will still only result in 2 records (second import a no-op).
        self.assertEqual(2, ExampleDeserializeConfig.objects.count())

        # Change Betty.
        betty = ExampleDeserializeConfig.current('betty')
        betty.int_field = -8
        betty.save()

        self.assertEqual(3, ExampleDeserializeConfig.objects.count())
        self.assertEqual(-8, ExampleDeserializeConfig.current('betty').int_field)

        # Now importing will add a new entry for Betty.
        with open(self.fixture_path, "rb") as data:
            entries_created = deserialize_json(data, self.test_username)
            self.assertEqual(1, entries_created)

        self.assertEqual(4, ExampleDeserializeConfig.objects.count())
        self.assertEqual(5, ExampleDeserializeConfig.current('betty').int_field)

    def test_bad_username(self):
        """
        Tests the error handling when the specified user does not exist.
        """
        test_json = textwrap.dedent("""
            {
                "model": "example.ExampleDeserializeConfig",
                "data": [{"name": "dino"}]
            }
            """).encode('utf-8')
        with self.assertRaisesRegex(Exception, "User matching query does not exist"):
            deserialize_json(io.BytesIO(test_json), "unknown_username")

    def test_invalid_json(self):
        """
        Tests the error handling when there is invalid JSON.
        """
        test_json = textwrap.dedent("""
            {
                "model": "config_models.ExampleDeserializeConfig",
                "data": [{"name": "dino"
            """).encode('utf-8')
        with self.assertRaisesRegex(Exception, "JSON parse error"):
            deserialize_json(io.BytesIO(test_json), self.test_username)

    def test_invalid_model(self):
        """
        Tests the error handling when the configuration model specified does not exist.
        """
        test_json = textwrap.dedent("""
            {
                "model": "xxx.yyy",
                "data":[{"name": "dino"}]
            }
            """).encode('utf-8')
        with self.assertRaisesRegex(Exception, "No installed app"):
            deserialize_json(io.BytesIO(test_json), self.test_username)


class PopulateModelTestCase(CacheIsolationTestCase):
    """
    Tests of populate model management command.
    """
    def setUp(self):
        super().setUp()
        self.file_path = os.path.join(os.path.dirname(__file__), 'data', 'data.json')
        self.test_username = 'test_management_worker'
        User.objects.create_user(username=self.test_username)

    def test_run_command(self):
        """
        Tests the "happy path", where 2 instances of the test model should be created.
        A valid username is supplied for the operation.
        """
        _run_command(file=self.file_path, username=self.test_username)
        self.assertEqual(2, ExampleDeserializeConfig.objects.count())

        betty = ExampleDeserializeConfig.current('betty')
        self.assertEqual(self.test_username, betty.changed_by.username)

        fred = ExampleDeserializeConfig.current('fred')
        self.assertEqual(self.test_username, fred.changed_by.username)

    def test_no_user_specified(self):
        """
        Tests that a username must be specified.
        """
        with self.assertRaisesRegex(CommandError, "A valid username must be specified"):
            _run_command(file=self.file_path)

    def test_bad_user_specified(self):
        """
        Tests that a username must be specified.
        """
        with self.assertRaisesRegex(Exception, "User matching query does not exist"):
            _run_command(file=self.file_path, username="does_not_exist")

    def test_no_file_specified(self):
        """
        Tests the error handling when no JSON file is supplied.
        """
        with self.assertRaisesRegex(CommandError, "A file containing JSON must be specified"):
            _run_command(username=self.test_username)

    def test_bad_file_specified(self):
        """
        Tests the error handling when the path to the JSON file is incorrect.
        """
        with self.assertRaisesRegex(CommandError, "File does/not/exist.json does not exist"):
            _run_command(file="does/not/exist.json", username=self.test_username)


def _run_command(*args, **kwargs):
    """Run the management command to deserializer JSON ConfigurationModel data. """
    command = populate_model.Command()
    return command.handle(*args, **kwargs)
